#define PYGGI_MODULE
#include "pyggi.h"

static int       Visual_default_setattr(Visual *, char *, PyObject *);
static PyObject* Visual_default_getattr(Visual *, char *);


#define VISUAL_FUNC(s, c) PyCFunction_New(&Visual_methods[c], (PyObject*)s)

static PyObject*
Visual_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
	Visual *self;
	
	self = (Visual*)type->tp_alloc(type, 0);
	if (self!=NULL) {
		self->vis     = NULL;
		self->pf      = NULL;
		self->exts    = NULL;
		self->nb_exts = 0;
		memset(&(self->mode), 0, sizeof(ggi_mode));

	}
	return (PyObject*)self;
}

static int
Visual_init(Visual * self, PyObject *args, PyObject *kwds) {
	char *a;

	/* XXX parse kwds */

	a = NULL;
	if(ggiInit()<0) {
		PyErr_SetString(PyExc_SystemError,"ggiInit failed");
		return -1;
	}
	if(!PyArg_ParseTuple(args, "|s", &a)) {
		PyErr_SetString(PyExc_TypeError,"string required");
		return -1;
	}
	self->vis = ggiOpen(a);
	if (self->vis == NULL) {
		PyErr_SetString(PyExc_SystemError,"ggiOpen failed");
		return -1;
	}
	
	ggiGetMode(self->vis, &(self->mode));
        
	/* TODO : Doesn't it depends on the mode ? */
        
	self->pf = ggiGetPixelFormat(self->vis);
	
	return 0;
}

static void
Visual_dealloc(Visual *self) {
	if(self->vis) {
		pyggi_detachAll(self);
		ggiClose(self->vis);
	}
	ggiExit();
	self->ob_type->tp_free((PyObject*)self);
}

static int
Visual_setattr(Visual *self, char *name, PyObject *value) {
	Extension *e;
	int r;
	
	/* Priority handlers  */
	
	/* Search in extensions */
	for(e = self->exts; e; e = e->next) {
		r = e->setattr(self, name, value);
		if(r==1)
			return 0;
		else if(r==-1)
			return -1;
	}
	
	/* Fallback to default handler */
	
	r = Visual_default_setattr(self, name, value);
	if(r==1)
		return 0;
	else if(r==-1)
		return -1;
	
	/* Not found */	
	
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

static PyObject*
Visual_getattr(Visual *self, char *name) {
	Extension *e;
	PyObject * r;
	
	/* Priority handlers  */	
	if(name[0]=='e' && !strcmp(name,"extensions")) {
		return pyggi_getExtensionList(self);
	}
	
	/* Search in extensions */
	for(e = self->exts; e; e = e->next) {
		r = e->getattr(self, name);
		if(r || PyErr_Occurred())
			return r;
	}
	
	/* Fallback to default handler */
	r = Visual_default_getattr(self, name);
	if(r || PyErr_Occurred())
		return r;
	
	/* Not found */

	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

static PyObject*
Visual_str(Visual * self) {
	return PyString_FromFormat("<ggi_visual_t at %p>",self->vis);
}

static PyObject*
Visual_repr(Visual * self) {
	return PyString_FromString("Visual()");
}


PyTypeObject VisualType = {
	PyObject_HEAD_INIT(NULL)
	0,                           /* ob_size         */
	"GGI._giimodule.Visual",     /* tp_name         */
	sizeof(Visual),              /* tp_basicsize    */
	0,                           /* tp_itemsize     */
	(destructor)Visual_dealloc,  /* tp_dealloc      */
	0,                           /* tp_print        */
	(getattrfunc)Visual_getattr, /* tp_getattr      */
	(setattrfunc)Visual_setattr, /* tp_setattr      */
	0,                           /* tp_compare      */
	(reprfunc)Visual_repr,       /* tp_repr         */
	0,                           /* tp_as_number    */
	0,                           /* tp_as_sequence  */
	0,                           /* tp_as_mapping   */
	0,                           /* tp_hash         */
	0,                           /* tp_call         */
	(reprfunc)Visual_str,        /* tp_str          */
	0,                           /* tp_getattro     */
	0,                           /* tp_setattro     */
	0,                           /* tp_as_buffer    */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,     /*tp_flags*/
	"Visual objects",            /* tp_doc          */
	
	/* 2.0 */
	0,		             /* tp_traverse     */
	0,		             /* tp_clear        */
	/* 2.1 */
	0,		             /* tp_richcompare  */
	0,		             /* tp_weaklistoffset */
	/* 2.2 */
	0,		             /* tp_iter         */
	0,		             /* tp_iternext     */
	0,                           /* tp_methods      */
	0,                           /* tp_members      */
	0,                           /* tp_getset       */
	0,                           /* tp_base         */
	0,                           /* tp_dict         */
	0,                           /* tp_descr_get    */
	0,                           /* tp_descr_set    */
	0,                           /* tp_dictoffset   */
	(initproc)Visual_init,       /* tp_init         */
	0,                           /* tp_alloc        */
	Visual_new,                  /* tp_new          */
	0,                           /* tp_is_gc        */
	0,                           /* tp_bases        */
        0,                           /* tp_mro          */
	0,                           /* tp_cache        */
	0,                           /* tp_subclasses   */
	0,                           /* tp_weaklist     */
	0                            /* tp_del          */
};



/******************************************************************
 * Input handling 
 ******************************************************************/

static PyObject*
Visual_join(Visual * self, PyObject* args) {
	PyObject * inp;

	if(!PyArg_ParseTuple(args, "O", &inp)) {
		PyErr_SetString(PyExc_TypeError,"Input required");
		return NULL;
	}
	
	ASSERT_INPUT(inp, NULL);
	ASSERT_VALID_INPUT(((Input*)inp), NULL);

	ggiJoinInputs(self->vis,((Input*)args)->inp);
	((Input*)inp)->inp=NULL; /* invalidate the object */
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_detach(Visual * self) {
	gii_input_t inp;
	Input *ret;
	
	if(self->vis->input==NULL) {
		PyErr_SetString(PyExc_TypeError,"not a valid input");
		return NULL;
	}
	
	inp = ggiDetachInput(self->vis);
	ret = PyObject_New(Input, gii_c_api->InputType);
	if(!ret) return NULL;
	ret->inp = inp;
	return (PyObject*)ret;
}



/******************************************************************
 * Event management
 ******************************************************************/

static PyObject*
Visual_read(Visual *self, PyObject *args, PyObject * kwds) {
	int mask;
	gii_event e;
	gii_event *ev;
	PyObject * d;
	static char *kwlist[]={ "mask", "event", NULL };
	
	if(self->vis->input==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}
	
	mask = emAll; /* by default */
	ev   = &e;
	d    = NULL;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|iO",
					kwlist, &mask, &d)) {
		return NULL;
	}
	if (d) {
		ASSERT_EVENT(d,NULL);
		ev=&(((Event*)(d))->ev);
	}
	
	if(ggiEventsQueued(self->vis,mask)) {
		ggiEventRead(self->vis, ev, mask);
		if (d==NULL)
			return gii_c_api->newEventFrom(e);	
		else
			/* XXX do we need to drop a reference? */
			return d;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_poll(Visual *self, PyObject*args, PyObject * kwds) {
	struct timeval t;
	struct timeval * tv;
	double f=-1.0;
	gii_event_mask mask;
	static char *kwlist[]={ "timeout", "mask", NULL };
	
	mask      = emAll;
	f         = -1.0;
	t.tv_sec  = 0;
	t.tv_usec = 0;
	tv        = NULL;
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|di",
					kwlist, &f, &mask)) {
		return NULL;
	}
	
	if (f>=0) {
		FLOAT_TO_TS(&t,f);
		tv=&t;
	}
	mask = ggiEventPoll(self->vis, mask,tv);
	if (mask==0) {
		gii_c_api->raiseTimeout();
		return NULL;
	}
	if (tv==NULL) /* no timeout */ 
		return PyInt_FromLong((long)mask);	
	return Py_BuildValue("id", mask, TS_TO_FLOAT(&t));
}


static PyObject*
Visual_queued(Visual *self, PyObject*args, PyObject * kwds) {
	int mask;
	static char *kwlist[]={ "mask", NULL };
	
	if(self->vis->input==NULL) {
		PyErr_SetString(PyExc_TypeError,"not a valid input");
		return NULL;
	}
	
	mask = emAll; /* by default */
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|i", kwlist, &mask))
		return NULL;
	return PyInt_FromLong((long)ggiEventsQueued(self->vis, mask));
}


static PyObject*
Visual_send(Visual *self, PyObject* args) {
	PyObject * e;

	if(self->vis->input==NULL) {
		PyErr_SetString(PyExc_TypeError,"Input invalid");
		return NULL;
	}
	
	if(!PyArg_ParseTuple(args, "O", &e)) {
		PyErr_SetString(PyExc_TypeError, "Event required");
		return NULL;
	}

	ASSERT_EVENT(e, NULL);
	
	if(ggiEventSend(self->vis, &(((Event*)e)->ev))<0) {
		PyErr_SetString(PyExc_SystemError,"ggiEventSend failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

/******************************************************************
 * Blitting
 ******************************************************************/

static PyObject*
Visual_copyBox(Visual *self, PyObject *args) {
	int x,y,w,h,nx,ny;

	switch(PyTuple_GET_SIZE(args)) {
	case 0:
		if(ggiCopyBox(self->vis,0,0,
			      self->mode.virt.x, self->mode.virt.y,0,0)) {
			PyErr_SetString(PyExc_SystemError,"ggiCopyBox failed");
			return NULL;
		}
		break;
	case 4:
		if(!PyArg_ParseTuple(args, "iiii",&x,&y,&w,&h))
			return NULL;
		if(ggiCopyBox(self->vis,x,y,w,h,x,y)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiCopyBox failed");
			return NULL;
		}
		break;
	case 6:
		if(!PyArg_ParseTuple(args, "iiiiii",&x,&y,&w,&h,&nx,&ny))
			return NULL;
		if(ggiCopyBox(self->vis,x,y,w,h,nx,ny)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiCopyBox failed");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError,"wrong parameters");
			return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_crossBlit(Visual *self, PyObject*args) {
	int x,y,w,h,nx,ny;
	Visual *o;
	
	switch(PyTuple_GET_SIZE(args)) {
	case 1:
		if(!PyArg_ParseTuple(args, "O",&o))
			return NULL;
		ASSERT_VISUAL(o, NULL);
		ASSERT_VALID_VISUAL(o, NULL);

		if(ggiCrossBlit(self->vis,0,0,
				self->mode.virt.x,
				self->mode.virt.y,o->vis,0,0)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiCrossBlit failed");
			return NULL;
		}
		break;
	case 5:
		if(!PyArg_ParseTuple(args, "Oiiii",&o,&x,&y,&w,&h))
			return NULL;

		ASSERT_VISUAL(o, NULL);
		ASSERT_VALID_VISUAL(o, NULL);

		if(ggiCrossBlit(self->vis,x,y,w,h,o->vis,x,y)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiCrossBlit failed");
			return NULL;
		}
		break;
	case 7:
		if(!PyArg_ParseTuple(args, "Oiiiiii",&o,&x,&y,&w,&h,&nx,&ny))
			return NULL;
		
		ASSERT_VISUAL(o, NULL);
		ASSERT_VALID_VISUAL(o, NULL);

		if(ggiCrossBlit(self->vis,x,y,w,h,o->vis,nx,ny)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiCrossBlit failed");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError,"wrong parameters");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}



/******************************************************************
 * Drawing
 ******************************************************************/

static PyObject*
Visual_putPixel(Visual *self, PyObject *args) {
	int x,y;
	ggi_pixel p;
	PyObject * o;
	
	switch(PyTuple_GET_SIZE(args)) {
	case 2:
		if(!PyArg_ParseTuple(args, "ii",&x,&y))
			return NULL;
		if(ggiDrawPixel(self->vis,x,y)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawPixel failed");
			return NULL;
		}
		break;
	case 3:
		if(!PyArg_ParseTuple(args, "iiO",&x,&y,&o))
			return NULL;
		if(pyggi_getPixel(o,&p,self->vis))
			return NULL;
		if(ggiPutPixel(self->vis,x,y,p)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiPutPixel failed");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError,"wrong parameters");
			return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_putLine(Visual *self, PyObject *args) {
	int _x1,_y1,_x2,_y2;
	ggi_pixel fg;
	ggi_pixel p;
	PyObject * o;

	switch(PyTuple_GET_SIZE(args)) {
	case 4:
		if(!PyArg_ParseTuple(args, "iiii",&_x1,&_y1,&_x2,&_y2))
			return NULL;
		if(ggiDrawLine(self->vis,_x1,_y1,_x2,_y2)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawLine failed");
			return NULL;
		}
		break;
	case 5:
		if(!PyArg_ParseTuple(args, "iiiiO",&_x1,&_y1,&_x2,&_y2,&o))
			return NULL;
		if(pyggi_getPixel(o,&p,self->vis))
			return NULL;
		ggiGetGCForeground(self->vis,&fg);
		ggiSetGCForeground(self->vis,p);
		_x1 = ggiDrawLine(self->vis,_x1,_y1,_x2,_y2);
		ggiSetGCForeground(self->vis,fg);
		if (_x1) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawLine failed");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError,"wrong parameters");
			return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_putHLine(Visual *self, PyObject *args) {
	int _x1,_y1;
	size_t l;
	ggi_pixel fg;
	ggi_pixel p;
	PyObject * o;

	switch(PyTuple_GET_SIZE(args)) {
	case 3:
		if(!PyArg_ParseTuple(args, "iii",&_x1,&_y1,&l))
			return NULL;
		if(ggiDrawHLine(self->vis,_x1,_y1,(int)l)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawHLine failed");
			return NULL;
		}
		break;
	case 4:
		if(!PyArg_ParseTuple(args, "iiiO",&_x1,&_y1,&l,&o))
			return NULL;
		/* test if o is a Pack */
		if PackCheck(o) {
			if PACK_HOLDS(o,l) {
				if(ggiPutHLine(self->vis,_x1,_y1,(int)l,
					       ((Pack *)o)->pixels)) {
					PyErr_SetString(PyExc_SystemError,
							"ggiPutHLine failed");
					return NULL;
				}
				break;
			}
			PyErr_SetString(PyExc_ValueError,
					"Pack is too small...");
			return NULL;
		}
		if (pyggi_getPixel(o,&p,self->vis))
			return NULL;
		ggiGetGCForeground(self->vis,&fg);
		ggiSetGCForeground(self->vis,p);
		_x1 = ggiDrawHLine(self->vis,_x1,_y1,(int)l);
		ggiSetGCForeground(self->vis,fg);
		if (_x1) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawHLine failed");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError,"wrong parameters");
			return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject*
Visual_putVLine(Visual *self, PyObject *args) {
	int _x1,_y1;
	size_t l;
	ggi_pixel fg;
	ggi_pixel p;
	PyObject * o;
	
	switch(PyTuple_GET_SIZE(args)) {
	case 3:
		if(!PyArg_ParseTuple(args, "iii",&_x1,&_y1,&l))
			return NULL;
		if(ggiDrawVLine(self->vis,_x1,_y1,(int)l)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawVLine failed");
			return NULL;
		}
		break;
	case 4:
		if(!PyArg_ParseTuple(args, "iiiO",&_x1,&_y1,&l,&o))
			return NULL;
		/* test if o is a Pack */
		if PackCheck(o) {
			if PACK_HOLDS(o,l) {
				if(ggiPutVLine(self->vis,_x1,_y1,(int)l,
					       ((Pack*)o)->pixels)) {
					PyErr_SetString(PyExc_SystemError,
							"ggiPutVLine failed");
					return NULL;
				}
				break;
			}
			PyErr_SetString(PyExc_ValueError,
					"Pack is too small...");
			return NULL;
		}
		if(pyggi_getPixel(o,&p,self->vis))
			return NULL;
		ggiGetGCForeground(self->vis,&fg);
		ggiSetGCForeground(self->vis,p);
		_x1 = ggiDrawVLine(self->vis,_x1,_y1,(int)l);
		ggiSetGCForeground(self->vis,fg);
		if(_x1) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawVLine failed");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError,"wrong parameters");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_putBox(Visual *self, PyObject *args) {
	int _x1,_y1,_w,_h;
	ggi_pixel fg;
	ggi_pixel p;
	PyObject * o;

	switch(PyTuple_GET_SIZE(args)) {
	case 4:
		if(!PyArg_ParseTuple(args, "iiii",&_x1,&_y1,&_w,&_h))
			return NULL;
		if(ggiDrawBox(self->vis,_x1,_y1,_w,_h)) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawBox failed");
			return NULL;
		}
		break;
	case 5:
		if(!PyArg_ParseTuple(args, "iiiiO",&_x1,&_y1,&_w,&_h,&o))
			return NULL;
		if PackCheck(o) {
			if PACK_HOLDS(o,(size_t)(_w*_h)) {
				if(ggiPutBox(self->vis,_x1,_y1,_w, _h,
					     ((Pack*)o)->pixels)) {
					PyErr_SetString(PyExc_SystemError,
							"ggiPutBox failed");
					return NULL;
				}
				break;
			}
			PyErr_SetString(PyExc_ValueError,
					"Pack is too small...");
			return NULL;
		}
		if(pyggi_getPixel(o,&p,self->vis))
			return NULL;
		ggiGetGCForeground(self->vis,&fg);
		ggiSetGCForeground(self->vis,p);
		_x1 = ggiDrawBox(self->vis,_x1,_y1,_w,_h);
		ggiSetGCForeground(self->vis, fg);
		if (_x1) {
			PyErr_SetString(PyExc_SystemError,
					"ggiDrawBox failed");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError,
				"wrong parameters");
			return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_putString(Visual *self, PyObject *args) {
	int x,y;
	char * s;
	
	if(!PyArg_ParseTuple(args, "iis",&x,&y,&s))
		return NULL;
	if(!ggiPuts(self->vis, x,y,s)) {
		PyErr_SetString(PyExc_SystemError, "ggiPuts failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}


/* Get */

static PyObject*
Visual_getPixel(Visual *self, PyObject *args) {
	int x,y;
	ggi_pixel p;
	
	if(!PyArg_ParseTuple(args, "ii",&x,&y))
		return NULL;
	
	if (ggiGetPixel(self->vis,x,y,&p) < 0) {
		PyErr_SetString(PyExc_SystemError,
				"ggiGetPixel failed");
		return NULL;
	}
	return PyInt_FromLong((long)p);
}

static PyObject*
Visual_getBox(Visual *self, PyObject *args) {
	int _x1,_y1,_w,_h;
	Pack * r;
	
	switch(PyTuple_GET_SIZE(args)) {
	case 4:
		if(!PyArg_ParseTuple(args, "iiii",&_x1,&_y1,&_w,&_h))
			return NULL;
		r = (Pack*)pyggi_newPackFromLength(self, (size_t)(_w*_h));
		if(!r) return NULL;
		break;
	case 5:
		if(!PyArg_ParseTuple(args, "iiiiO",&_x1,&_y1,&_w,&_h,&r))
			return NULL;
		if PackCheck(r) {
			if PACK_HOLDS(r, (size_t)(_w*_h)) {
				/* drop a reference? */
				break;
			}
			PyErr_SetString(PyExc_ValueError,
					"Pack is too small...");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError, "wrong parameters");
		return NULL;
	}
	if (ggiGetBox(self->vis,_x1,_y1,_w, _h,r->pixels)) {
		PyErr_SetString(PyExc_SystemError, "ggiGetBox failed");
		/* XXX free r */
		return NULL;
	}
	return (PyObject*)r;
}

static PyObject*
Visual_getHLine(Visual *self, PyObject *args) {
	int _x1,_y1;
	size_t l;
	Pack * r;

	switch(PyTuple_GET_SIZE(args)) {
	case 4:
		if(!PyArg_ParseTuple(args, "iii",&_x1,&_y1,&l))
			return NULL;
		r = (Pack*)pyggi_newPackFromLength(self, l);
		if(!r) return NULL;
		break;
	case 5:
		if(!PyArg_ParseTuple(args, "iiiO",&_x1,&_y1,&l,&r))
			return NULL;
		if PackCheck(r) {
			if PACK_HOLDS(r,l) {
				/* drop a reference? */
				break;
			}
			PyErr_SetString(PyExc_ValueError,
					"Pack is too small...");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError, "wrong parameters");
		return NULL;
	}
	if (ggiGetHLine(self->vis, _x1, _y1, (int)l,r->pixels)) {
		PyErr_SetString(PyExc_SystemError, "ggiGetHLine failed");
		/* XXX free r */
		return NULL;
	}
	return (PyObject*)r;
}

static PyObject*
Visual_getVLine(Visual *self, PyObject *args) {
	int _x1,_y1;
	size_t l;
	Pack * r;

	switch(PyTuple_GET_SIZE(args)) {
	case 4:
		if(!PyArg_ParseTuple(args, "iii",&_x1,&_y1,&l))
			return NULL;
		r = (Pack*)pyggi_newPackFromLength(self, l);
		if(!r) return NULL;
		break;
	case 5:
		if(!PyArg_ParseTuple(args, "iiiO",&_x1,&_y1,&l,&r))
			return NULL;
		if PackCheck(r) {
			if PACK_HOLDS(r,l) {
				/* drop a reference? */
				break;
			}
			PyErr_SetString(PyExc_ValueError,
					"Pack is too small...");
			return NULL;
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError, "wrong parameters");
		return NULL;
	}
	if(ggiGetVLine(self->vis, _x1, _y1, (int)l,r->pixels)) {
		PyErr_SetString(PyExc_SystemError, "ggiGetVLine failed");
		/* XXX free r */
		return NULL;
	}
	return (PyObject*)r;
}

static PyObject*
Visual_Pack(Visual *self, PyObject *args) {
	PyObject * o;

	switch PyTuple_GET_SIZE(args) {
	case 1:
		o = PyTuple_GET_ITEM(args,0);
		
		/* 1. the arg is an integer 
		   -> allocate a pack of the given size */
		if(PyInt_Check(o))
			return pyggi_newPackFromLength(self, PyInt_AsLong(o));
		
		/* 2. the arg is an array of float triplets
		   -> allocate a pack and ggiPack these */
		if(PySequence_Check(o)) {
			return pyggi_newPackFromSequence(self, o);
		}
		
		/* 3. the arg is a pack 
		   -> allocate a new pack and copy from arg, with
		   conversion is the visual are different */		
		if(PackCheck(o)) {
			return pyggi_newPackFromPack(self, (Pack*)o);
		}
		break;
	default:
		PyErr_SetString(PyExc_TypeError, "Need at most one argument");
		return NULL;
	}
	PyErr_SetString(PyExc_TypeError, "Need at most one argument");
	return NULL;
}

static PyObject*
Visual_Mode(Visual *self, PyObject *args) {
	char * s;
	ggi_mode m;
	
	s = "";
	if(!PyArg_ParseTuple(args, "|s",&s))
		return NULL;
	ggiParseMode(s, &m);
	ggiCheckMode(self->vis, &m);
	return pyggi_newMode(m);
}

static PyObject*
Visual_map(Visual *self, PyObject *args) {
	float r,g,b;
	ggi_color col;

	if(!PyArg_ParseTuple(args, "fff",&r,&g,&b))
		return NULL;
	float2color(r,g,b,&col);
	return PyInt_FromLong((long)ggiMapColor(self->vis,&col));
}

static PyObject*
Visual_unMap(Visual *self, PyObject *args) {
	ggi_pixel p;
	double fa,fb,fc;
	ggi_color col;
	
	if(!PyArg_ParseTuple(args, "i",&p))
		return NULL;
	ggiUnmapPixel(self->vis, p, &col);
	color2float(&col,fa,fb,fc);
	return Py_BuildValue("[fff]", fa,fb,fc);
}

/******************************************************************
 * Flushing
 ******************************************************************/

static PyObject*
Visual_flush(Visual *self, PyObject *args) {
	int _x1,_y1,_x2,_y2;
	
	if(PyTuple_GET_SIZE(args)==0) 
		ggiFlush(self->vis);
	else {
		if(!PyArg_ParseTuple(args, "iiii", &_x1, &_y1, &_x2, &_y2))
			return NULL;
		ggiFlushRegion(self->vis,
			       _x1,_y1, (_x2 - _x1 + 1), (_y2 - _y1 + 1));
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject*
Visual_fill(Visual*self, PyObject *args) {
	ggi_pixel p, fg;
	switch(PyTuple_GET_SIZE(args)) {
	case 0:
		ggiFillscreen(self->vis);
		Py_INCREF(Py_None);
		return Py_None;
	case 1:
		if(pyggi_getPixel(PyTuple_GET_ITEM(args,0), &p, self->vis)<0) {
			return NULL;
		}
		ggiGetGCForeground(self->vis, &fg);
		ggiSetGCForeground(self->vis, p);
		ggiFillscreen(self->vis);
		ggiSetGCForeground(self->vis, fg);
		Py_INCREF(Py_None);
		return Py_None;
	default:
		break;
	}
	PyErr_SetString(PyExc_TypeError, "Need at most one argument");
	return NULL;
}


/****************************************************************/

enum {
	VISUAL_PUTPIXEL,
	VISUAL_PUTLINE,
	VISUAL_PUTHLINE,
	VISUAL_PUTVLINE,
	VISUAL_PUTBOX,

	VISUAL_PUTSTRING,
	
	VISUAL_GETPIXEL,
	VISUAL_GETHLINE,
	VISUAL_GETVLINE,
	VISUAL_GETBOX,

	VISUAL_CROSSBLIT,
	VISUAL_COPYBOX,
	
	VISUAL_MAP,
	VISUAL_UNMAP,	

	/*
	  VISUAL_SETPALETTE
	  VISUAL_GETPALETTE

	  VISUAL_SETGAMMAMAP
	  VISUAL_GETGAMMAMAP
	*/
	


	VISUAL_PACK,
	
	VISUAL_MODE,
	VISUAL_FILL,
	VISUAL_FLUSH,

	VISUAL_JOIN,
	VISUAL_DETACH,
	VISUAL_READ,
	VISUAL_POLL,
	VISUAL_QUEUED,
	VISUAL_SEND,
};

static struct PyMethodDef Visual_methods[] = {
	
	/* ---------- drawing ops --------------------*/
	
	{ "putPixel", (PyCFunction)Visual_putPixel, METH_VARARGS,(char*)NULL},
	{ "putLine",  (PyCFunction)Visual_putLine,  METH_VARARGS,(char*)NULL},
	{ "putHLine", (PyCFunction)Visual_putHLine, METH_VARARGS,(char*)NULL},
	{ "putVLine", (PyCFunction)Visual_putVLine, METH_VARARGS,(char*)NULL},
	{ "putBox",   (PyCFunction)Visual_putBox,   METH_VARARGS,(char*)NULL},
	
	{ "putString",(PyCFunction)Visual_putString,METH_VARARGS,(char*)NULL},

	/* ----------- getops  */
	
	{ "getPixel", (PyCFunction)Visual_getPixel, METH_VARARGS,(char*)NULL},
	{ "getHLine", (PyCFunction)Visual_getHLine, METH_VARARGS,(char*)NULL},
	{ "getVLine", (PyCFunction)Visual_getVLine, METH_VARARGS,(char*)NULL},
	{ "getBox",   (PyCFunction)Visual_getBox,   METH_VARARGS,(char*)NULL},
	
	/* ----------- blitting ops -----------------*/
	
	{ "crossBlit",(PyCFunction)Visual_crossBlit,METH_VARARGS,(char*)NULL},
	{ "copyBox",  (PyCFunction)Visual_copyBox,  METH_VARARGS,(char*)NULL},

	/* ------------- color conversion ------------*/

	{ "map",      (PyCFunction)Visual_map,      METH_VARARGS,(char*)NULL},
	{ "unMap",    (PyCFunction)Visual_unMap,    METH_VARARGS,(char*)NULL},
	
	/* ---------- packs ------------------------ */
	
	{ "Pack",     (PyCFunction)Visual_Pack,     METH_VARARGS,(char*)NULL},

	/* ---------- visual ops --------------------*/

	{ "Mode",     (PyCFunction)Visual_Mode,     METH_VARARGS, (char*)NULL},
	{ "fill",     (PyCFunction)Visual_fill,     METH_VARARGS, (char*)NULL},
	{ "flush",    (PyCFunction)Visual_flush,    METH_VARARGS, (char*)NULL},
	
	/* ---------- Input management ----------------*/

	{ "join",    (PyCFunction)Visual_join,    METH_VARARGS, (char*)NULL},
	{ "detach",  (PyCFunction)Visual_detach,  METH_NOARGS,  (char*)NULL},

	/* ------------- Events functions ----------------------- */

	{ "read",  (PyCFunction)Visual_read,  METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL },
	{ "poll",  (PyCFunction)Visual_poll,  METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL },
	{ "queued",(PyCFunction)Visual_queued,METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL },
	{ "send",  (PyCFunction)Visual_send,  METH_VARARGS,  (char*)NULL},

	/* ------------ end ------------ */

	{ NULL, NULL, 0,(char*)NULL }
};

static PyObject *
Visual_default_getattr(Visual *self, char*name) {
	ggi_pixel p;
	ggi_mode mode;
	int a,b,c,d;
	ggi_float gr, gg, gb;
	
	switch(name[0]) {		
	case 'b':
		if(!strcmp(name,"bg")) {
			ggiGetGCBackground(self->vis,&p);
			return PyInt_FromLong((long)p);
		}
		break;
	case 'c':
		if(!strcmp(name,"clipping")) {
			ggiGetGCClipping(self->vis,&a,&b,&c,&d);
			return Py_BuildValue("iiii",a,b,c,d);
		}
		if(!strcmp(name,"copyBox"))
			return VISUAL_FUNC(self, VISUAL_COPYBOX);
		if(!strcmp(name,"crossBlit"))
			return VISUAL_FUNC(self, VISUAL_CROSSBLIT);
		if(!strcmp(name,"charSize")) {
			ggiGetCharSize(self->vis, &a, &b);
			return Py_BuildValue("ii",a,b);
		}
		break;
	case 'd':
		if(!strcmp(name,"displayFrame"))
			return PyInt_FromLong((long)
					      ggiGetDisplayFrame(self->vis));
		if(!strcmp(name,"devices"))
			return gii_c_api->getDeviceInfo(self->vis->input);
		if(!strcmp(name,"detach"))
			return VISUAL_FUNC(self, VISUAL_DETACH);
		break;
	case 'f':
		if(!strcmp(name,"flush"))
			return VISUAL_FUNC(self, VISUAL_FLUSH);
		if(!strcmp(name,"fill"))
			return VISUAL_FUNC(self, VISUAL_FILL);
		if(!strcmp(name,"flags"))
			return PyInt_FromLong((long)ggiGetFlags(self->vis));
		if(!strcmp(name,"fg")) {
			ggiGetGCForeground(self->vis,&p);
			return PyInt_FromLong((long)p);
		}
		break;
	case 'g':
		if(!strcmp(name,"gamma")) {
			ggiGetGamma(self->vis, &gr, &gg, &gb);
			return Py_BuildValue("fff",gr,gg,gb);
		}
		/* getGammaMap */
		if(!strcmp(name,"getPixel"))
			return VISUAL_FUNC(self, VISUAL_GETPIXEL);
		if(!strcmp(name,"getBox"))
			return VISUAL_FUNC(self, VISUAL_GETBOX);
		if(!strcmp(name,"getHLine"))
			return VISUAL_FUNC(self, VISUAL_GETHLINE);
		if(!strcmp(name,"getVLine"))
			return VISUAL_FUNC(self, VISUAL_GETVLINE);
		break;
	case 'j':
		if(!strcmp(name,"join"))
			return VISUAL_FUNC(self, VISUAL_JOIN);
		break;
	case 'm':
		if(!strcmp(name,"mask"))
			return PyInt_FromLong((long)
					      ggiGetEventMask(self->vis));
		if(!strcmp(name,"map"))
			return VISUAL_FUNC(self, VISUAL_MAP);
		
		if(!strcmp(name,"mode")) {
			ggiGetMode(self->vis,&mode);
			return pyggi_newMode(mode);
		}		
		break;
	case 'o':
		if(!strcmp(name,"origin")) {
			ggiGetOrigin(self->vis,&a,&b);
			return Py_BuildValue("ii",a,b);
		}
		break;
	case 'p': /* palette */
		if(!strcmp(name,"putPixel"))
			return VISUAL_FUNC(self, VISUAL_PUTPIXEL);
		if(!strcmp(name,"putBox"))
			return VISUAL_FUNC(self, VISUAL_PUTBOX);
		if(!strcmp(name,"putHLine"))
			return VISUAL_FUNC(self, VISUAL_PUTHLINE);
		if(!strcmp(name,"putVLine"))
			return VISUAL_FUNC(self, VISUAL_PUTVLINE);
		if(!strcmp(name,"putLine"))
			return VISUAL_FUNC(self, VISUAL_PUTLINE);
		if(!strcmp(name,"putString"))
			return VISUAL_FUNC(self, VISUAL_PUTSTRING);
		if(!strcmp(name,"poll"))
			return VISUAL_FUNC(self, VISUAL_POLL);		
		break;
	case 'q':
		if(!strcmp(name,"queued"))
			return VISUAL_FUNC(self, VISUAL_QUEUED);
		break;
	case 'r':
		if(!strcmp(name,"read"))
			return VISUAL_FUNC(self, VISUAL_READ);
		if(!strcmp(name,"readFrame"))
			return PyInt_FromLong((long)
					      ggiGetReadFrame(self->vis));
		break;
	case 's':
		if(!strcmp(name,"send"))
			return VISUAL_FUNC(self, VISUAL_SEND);
		/* setGammaMap */
		break;
		
	case 'u':
		if(!strcmp(name,"unMap"))
			return VISUAL_FUNC(self, VISUAL_UNMAP);
		break;
	case 'w':
		if(!strcmp(name,"writeFrame"))
			return PyInt_FromLong((long)
					      ggiGetWriteFrame(self->vis));
		break;
	case 'M':
		if(!strcmp(name,"Mode"))
			return VISUAL_FUNC(self, VISUAL_MODE);
		break;
	case 'P':
		if(!strcmp(name,"Pack"))
			return VISUAL_FUNC(self, VISUAL_PACK);
		break;
	}
	/* Not found */
	return NULL;
}

/*
  return 0 if not found 1 if ok -1 on error
 */
static int
Visual_default_setattr(Visual *self, char *name, PyObject *value) {
	int i;
	ggi_pixel p;
	ggi_coord c;
	int r[4];
	ggi_float gr, gg, gb;

	switch(name[0]) {
	case 'b':
		if(!strcmp(name,"bg")) {
			if(pyggi_getPixel(value, &p, self->vis)<0)
				return -1;
			ggiSetGCBackground(self->vis,p);
			return 1;
		}
		break;
	case 'c':
		if(!strcmp(name,"clipping")) {
			if(pyggi_getRect(value, r)<0)
				return -1;
			if(ggiSetGCClipping(self->vis,r[0],r[1],r[2],r[3])) {
				PyErr_SetString(PyExc_SystemError,
						"ggiSetGCClipping failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'd':
		if(!strcmp(name,"displayFrame")) {
			if(pyggi_getInt(value, &i)<0)
				return -1;
			if(ggiSetDisplayFrame(self->vis,i)) {
				PyErr_SetString(PyExc_SystemError,
						"ggiSetDisplayFrame failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'f':
		if(!strcmp(name,"fg")) {
			if(pyggi_getPixel(value, &p,self->vis)<0)
				return -1;
			ggiSetGCForeground(self->vis,p);
			return 1;
		}
		if(!strcmp(name,"flags")) {
			if(pyggi_getInt(value, &i)<0)
				return -1;
			ggiSetFlags(self->vis, (ggi_flags)i);
			return 1;
		}
		break;
	case 'g':
		if(!strcmp(name,"gamma")) {
			if(!PyArg_ParseTuple(value, "fff",&gr,&gg,&gb))
				return -1;
			ggiSetGamma(self->vis, gr, gg, gb);
			return 1;
		}
		break;
	case 'm':
		if(!strcmp(name,"mask")) {
			if(pyggi_getInt(value, &i)<0)
				return -1;
			ggiSetEventMask(self->vis, (gii_event_mask)i);
			return 1;
		}
		if(!strcmp(name,"mode")) {
			if (!ModeCheck(value)) {
				PyErr_SetString(PyExc_TypeError,
						"Mode expected");
				return -1;
			}
			ggiSetMode(self->vis, &((Mode*)value)->mode);
			ggiGetMode(self->vis,&(self->mode));
			self->pf = ggiGetPixelFormat(self->vis);
			return 1;
		}
		break;
	case 'o':
		if(!strcmp(name,"origin")) {
			if(pyggi_getCoords(value, &c)<0)
				return -1;
			if(ggiSetOrigin(self->vis,c.x,c.y)) {
				PyErr_SetString(PyExc_SystemError,
						"ggiSetOrigin failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'p': /* palette */
		break;
	case 'r':
		if(!strcmp(name,"readFrame")) {
			if(pyggi_getInt(value, &i)<0)
				return -1;
			if(ggiSetReadFrame(self->vis,i)) {
				PyErr_SetString(PyExc_SystemError,
						"ggiSetReadFrame failed");
				return -1;
			}
			return 1;
		}
		break;
	case 'w':
		if(!strcmp(name,"writeFrame")) {
			if(pyggi_getInt(value, &i)<0)
				return -1;
			if(ggiSetWriteFrame(self->vis,i)) {
				PyErr_SetString(PyExc_SystemError,
						"ggiSetWriteFrame failed");
				return -1;
			}
			return 1;
		}
		break;
	}
	/* Not found */
	return 0;
}
